# Disk class
#
# Copyright 2006-2007 Novell, Inc.
# Author: Charles Coffing <ccoffing@novell.com>
#
# This software may be freely redistributed under the terms of the GNU
# general public license.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import os
import stat
import struct
import tempfile
import time

from vminstall.exceptions import *
from vminstall.log import log
import vminstall.msg
import vminstall.paths
import vminstall.util

def get_fv_boot_disks(disks):
    """Qemu can only boot from the first hard disk or first cdrom.
    Obviously the device must exist, and prefer CDROMs."""
    fv_boot_disks = []
    cdroms = filter(lambda disk: disk.vdevtype == Disk.DEVICE_CDROM, disks)
    hds = filter(lambda disk: disk.vdevtype == Disk.DEVICE_DISK, disks)
    if cdroms and cdroms[0].exists():
        fv_boot_disks.append(cdroms[0])
    if hds and hds[0].exists():
        fv_boot_disks.append(hds[0])
    return fv_boot_disks

def get_possible_vdevs(full_virt):
    def get_virtual_disks(base, start, stop):
        virtual_disks = []
        for x in [chr(i) for i in xrange(ord(start), ord(stop)+1)]:
            virtual_disks.append(base + x)
        return virtual_disks
    if full_virt:
        return get_virtual_disks('hd', 'a', 'd')
    else:
        return get_virtual_disks('xvd', 'a', 'p')

def get_unused_vdevs(full_virt, disks):
    all_vdevs = get_possible_vdevs(full_virt)
    used_vdevs = map(lambda d: d.vdev, disks)
    return filter(lambda v: v not in used_vdevs, all_vdevs)

def create_file(filename, bytes, args={}, progress=lambda f:None):
    if args.has_key('sparse'):
        sparse = args['sparse']
    else:
        sparse = 1
    log.debug("Creating raw file '%s', %d bytes (sparse=%d)." % (filename, bytes, sparse))
    fd = os.open(filename, os.O_WRONLY | os.O_CREAT | os.O_TRUNC)
    os.chmod(filename, 0600)
    if sparse:
        offset = bytes - 1
        os.lseek(fd, offset, 0)
        os.write(fd, '\x00')
    else:
        block_size = 65536
        data = '\x00' * block_size
        blocks = remaining = bytes / block_size
        block = 0
        while remaining:
            os.write(fd, data)
            remaining -= 1
            block += 1
            if (block & 1023) == 0:
                progress(float(block) / float(blocks))
    os.close(fd)
    progress(1.0)

def create_qcow(filename, bytes, arg={}, progress=lambda f:None):
    log.debug("Creating qcow file '%s', %d bytes." % (filename, bytes))
    mb = str(bytes / 1024 / 1024)
    vminstall.util.run(['/usr/sbin/qcow-create', mb, filename])

def devid_to_devstr(devid):
    return 'xvd' + (chr((devid-51712)/16 + ord('a')))

def block_attach(backdev, ro=True):
    mode = 'w'
    if ro:
        mode = 'r'
    for frontdev in xrange(51712, 51712+16*16, 16):
        try:
            vminstall.util.run(['/usr/sbin/xm', 'block-attach', '0', backdev, devid_to_devstr(frontdev), mode])
        except:
            continue
        # Unfortunately, xm returns before the hotplugging is guaranteed to
        # have been completed.  *sigh*
        # FIXME:  if this fixes it, put a stat/sleep loop
        time.sleep(1.0)
        return frontdev
    raise XenError(XenError.E_OTHER)

def block_detach(frontdev):
    # NOTE:  Some versions of Xen have a bug that affects detaching.  'xvda'
    #        may fail to detach, but '51712' works ok.  To be safe, the
    #        numeric form is always used.
    vminstall.util.run(['/usr/sbin/xm', 'block-detach', '0', str(frontdev), '-f'])

class Driver:
    def __init__(self, name):
        self.proto = name
    def vsize(self, path):
        return None
    def size(self, path):
        return None
    def exists(self, path):
        return True
    def is_local(self):
        return False
    def block_attach(self, disk):
        disk._block_vdev = block_attach(disk.get_pdev())
    def block_detach(self, disk):
        if disk._block_vdev:
            block_detach(disk._block_vdev)
            disk._block_vdev = None

class PhyDriver(Driver): 
    def vsize(self, path):
        return self.size(path)
    def size(self, path):
        try:
            f = open(path)
            f.seek(0, 2)  # seek to end
            return f.tell() / 512
        except:
            return None
    def is_local(self):
        return True
    def block_attach(self, disk):
        disk._block_vdev = disk.get_path()
    def block_detach(self, disk):
        disk._block_vdev = None

class FileDriver(Driver):
    def vsize(self, path):
        try:
            stats = os.stat(path)
            # FIXME:  kill this:
            # Treat 0-byte files as not existing (to allow use of mkstemp before
            # ever even creating a disk object, to avoid security races)
            return stats.st_size / 512 or None
        except:
            return None
    def size(self, path):
        try:
            stats = os.stat(path)
            return stats.st_blocks
        except:
            return None
    def exists(self, path):
        return os.path.exists(path)
    def is_local(self):
        return True

class LoopDriver(FileDriver):
    def create(self, path, blocks, args={}, progress=lambda f:None):
        dpath = os.path.dirname(path)
        if not os.path.exists(dpath):
            os.makedirs(dpath)
        create_file(path, blocks * 512, args, progress=progress)
    def block_attach(self, disk):
        # Raw files are already visible in the local filesystem; good enough.
        disk._block_vdev = disk.get_path()
    def block_detach(self, disk):
        disk._block_vdev = None

class TapAioDriver(FileDriver):
    def create(self, path, blocks, args={}, progress=lambda f:None):
        dpath = os.path.dirname(path)
        if not os.path.exists(dpath):
            os.makedirs(dpath)
        create_file(path, blocks * 512, args, progress=progress)

class TapVmdkDriver(FileDriver):
    def vsize(self, path):
        return None
    def is_local(self):
        return True

class TapQcowDriver(FileDriver):
    def create(self, path, blocks, args={}, progress=lambda f:None):
        dpath = os.path.dirname(path)
        if not os.path.exists(dpath):
            os.makedirs(dpath)
        create_qcow(path, blocks * 512, args, progress=progress)
    def vsize(self, path):
        # FIXME: how to query this?
        return None
    def is_local(self):
        return True

DRIVERS = {
    'file':     LoopDriver('file'),
    'iscsi':    Driver('iscsi'),
    'nbd':      Driver('nbd'),
    'npiv':     Driver('npiv'),
    'phy':      PhyDriver('phy'),
    'tap:qcow': TapQcowDriver('tap:qcow'),
    'tap:aio':  TapAioDriver('tap:aio'),
    'tap:vmdk': TapVmdkDriver('tap:vmdk'),
}
default_proto = 'file'

def parse_string(disk, extended=False, blocks=None):
    """Parses a text representation of a disk, and returns such a disk object.

    Command-line form:
    PDEV[,VDEV[,TYPE[,MODE[,MB[,OPTIONS]]]]]
    Extended form:
    PDEV,VDEV,TYPE,MODE,MB,DESCR
    """
    def bad():
        raise ParamError(err=ParamError.E_INVALID, details='VDEV')
    def escaped_split(s, sep=',', maxsplit=0, esc='\\'):
        words = ['']
        escaped = False
        for c in s:
            if c == esc and not escaped:
                escaped = True
                continue
            if c == sep and not escaped and (not maxsplit or len(words) < maxsplit+1):
                words.append('')
            else:
                words[-1] += c
            escaped = False
        return words
    params = escaped_split(disk, ',', 5)
    if extended:
        if len(params) < 6:
            bad()
    else:
        if len(params) < 1:
            bad()
    pdev = params[0]
    vdev = None
    vdevType = None
    ro = None
    if extended or len(params) > 1:
        vdev = params[1]
        if extended or len(params) > 2:
            vdevType = params[2]
            if extended or len(params) > 3:
                if params[3] not in ('r', 'w'):
                    bad()
                ro = (params[3] == 'r')
                if extended or len(params) > 4:
                    try:
                        blocks = int(float(params[4]) * 2*1024)
                    except:
                        bad()
    disk = Disk(pdev=pdev, vdev=vdev, vdevType=vdevType, ro=ro, blocks=blocks)
    if extended:
        disk.description = params[5]
    else:
        for arg in params[5:]:
            kv = arg.split('=', 1)
            if len(kv) == 1:
                kv[1] = True
            else:
                try:
                    kv[1] = int(kv[1])
                except:
                    pass
            disk.args[kv[0]] = kv[1]
    return disk

def split_driver_path(pdev):
    def guess_proto(path):
        proto = default_proto
        if os.path.exists(path):
            try:
                if stat.S_ISBLK(os.stat(path).st_mode):
                   proto = 'phy'
            except:
                pass
        return proto
    items = pdev.split(':', 1)
    if len(items) == 1:
        # Legacy support; no explicit driver
        path = items[0]
        proto = guess_proto(path)
    else:
        proto, path = items
        if proto == 'tap':
            subproto = path.split(':', 1)
            if len(subproto) == 2:
                proto = '%s:%s' % (proto, subproto[0])
                path = subproto[1]
    if proto not in DRIVERS:
        # Assume no protocol and that colon was part of pathname.  Undo everything.
        path = '%s:%s' % (proto, path)
        proto = guess_proto(path)
    return (DRIVERS[proto], path)

class Disk(object):
    DEVICE_DISK = 'disk'
    DEVICE_CDROM = 'cdrom'
    DEVICE_FLOPPY = 'floppy'

    translations = (
        (DEVICE_DISK, vminstall.msg.type_hard_disk),
        (DEVICE_CDROM, vminstall.msg.type_cdrom),
        (DEVICE_FLOPPY, vminstall.msg.type_floppy))

    def __init__(self, pdev, vdev=None, vdevType=None, ro=False, blocks=None,
                 full_virt=False, vmname='vm', disk_group=[], args={}):
        if vdevType is None:
            vdevType = self.DEVICE_DISK
        if vdevType not in (self.DEVICE_DISK, self.DEVICE_CDROM):
            raise ParamError(err=ParamError.E_INVALID, details='vdevType')

        if blocks is None:
            blocks = 2*1024*1024*4  # 4GB

        self._block_vdev = None
        self._template_lock = False
        self._created = False
        self.full_virt = full_virt
        self.vmname = vmname
        self.disk_group = disk_group
        self.description = ''
        
        self.set_pdev(pdev)
        self.set_vdev(vdev)
        self.set_vdevtype(vdevType)
        self.set_virtual_size(blocks)
        self.set_ro(ro)
        self.args = args

    def __del__(self):
        self.block_detach()

    def update(self, full_virt=None, vmname=None):
        if not self._template_lock:
            self._template_path = None
        if full_virt is not None:
           self.full_virt = full_virt
        if vmname is not None:
            self.vmname = vmname

    def copy(self):
        disk = Disk(pdev=self.get_pdev(), vdev=self._vdev_index, vdevType=self._vdevType,
                    ro=self._ro, blocks=self._blocks, args=self.args,
                    full_virt=self.full_virt) #FIXME??? , disk_group=self.disk_group)
        disk._template = self._template
        disk._template_lock = self._template_lock
        disk._template_path = self._template_path
        return disk

    def exists(self):
        return self.driver.exists(self.get_path())

    def block_attach(self):
        self.driver.block_attach(self)
        return self.attach_point()

    def block_detach(self):
        if 'driver' in dir(self):
            self.driver.block_detach(self)

    def attach_point(self):
        T = type(self._block_vdev)
        if T is int:
            return '/dev/' + devid_to_devstr(self._block_vdev)
        if T is str:
            return self._block_vdev
        assert(T is None)
        return None

    def has_bootsector(self):
        vdev = self.attach_point()
        assert(vdev is not None)
        pdev = self.get_path()
        log.info("Checking '%s' for bootsector..." % pdev)
        try:
            fd = os.open(vdev, os.O_RDONLY)
            buf = os.read(fd, 512)
        except:
            return False
        os.close(fd)
        if len(buf) == 512 and struct.unpack('H', buf[0x1fe: 0x200]) == (0xaa55,):
            log.info("Bootsector found on '%s'." % pdev)
            return True
        return False

    def is_template(self):
        return self._template

    def setup_time(self):
        """Returns a unit-less number, which attempts to describe the relative
        amount of time it will take to set up the disk."""
        if self.driver.proto != 'file' or self.exists():
            return 0.1
        if not self.args.has_key('sparse') or self.args['sparse']:
            return 1.0
        return self._blocks / (40.0*1024*1024)
    def setup(self, progress=lambda f:None):
        if self.driver.is_local() and self._path:
            self._path = os.path.abspath(self._path)
        self.validate()
        path = self.get_path()
        if 'create' in dir(self.driver) and not self.driver.size(path):
            log.info("Creating disk '%s'" % path)
            self.driver.create(path, self._blocks, self.args, progress)
            self._created = True
            self._template_lock = True
        progress(1.0)

    def reset(self):
        """If the disk was created, destroys the disk image and allows the disk
        to become a template again (if appropriate).  Existing disks are unchanged."""
        if self._created == True:
            self.delete()

    def delete(self):
        if 'create' not in dir(self.driver):
            raise NotImplementedError
        path = self.get_path()
        log.info("Deleting created disk '%s'" % path)
        os.unlink(path)
        self._created = False
        self._template_lock = False

    def validate(self):
        if self.vdevtype == self.DEVICE_CDROM and not self.exists():
            raise DiskResourceError(DiskResourceError.E_NOT_FOUND, details=self.get_pdev())
        if self._vdev_index >= len(get_possible_vdevs(self.full_virt)):
            raise ParamError(err=ParamError.E_INVALID, details=self.vdev)

    def get_vdev_index(self):
        return self._vdev_index
    def get_vdev_str(self):
        return get_possible_vdevs(self.full_virt)[self._vdev_index]
    def set_vdev(self, vdev):
        fvvdevs = get_possible_vdevs(True)
        pvvdevs = get_possible_vdevs(False)
        if vdev is None:
            unused_vdevs = get_unused_vdevs(self.full_virt, self.disk_group)
            if len(unused_vdevs) == 0:
                raise DiskResourceError(DiskResourceError.E_NOT_FOUND)
            else:
                vdev = unused_vdevs[0]
        ok = True
        if type(vdev) == str:
            if vdev in fvvdevs:
                self._vdev_index = fvvdevs.index(vdev)
            elif vdev in pvvdevs:
                self._vdev_index = pvvdevs.index(vdev)
            else:
                try:
                    self._vdev_index = int(vdev)
                except:
                    ok = False
        elif type(vdev) == int and vdev < max(len(fvvdevs), len(pvvdevs)):
            self._vdev_index = vdev
        else:
            ok = False
        if not ok:
            raise DiskResourceError(DiskResourceError.E_NOT_FOUND, details=vdev)
    vdev = property(get_vdev_str, set_vdev)

    def get_vdevtype(self):
        return self._vdevType
    def set_vdevtype(self, vdevtype):
        if not vdevtype in (self.DEVICE_DISK, self.DEVICE_CDROM, self.DEVICE_FLOPPY):
            raise ParamError(err=ParamError.E_INVALID, details='VDEV')
        self._vdevType = vdevtype
    vdevtype = property(get_vdevtype, set_vdevtype)
    def get_vdev_type_str(self):
        return filter(lambda row: row[0] == self._vdevType, self.translations)[0][1]

    def get_size(self):
        blocks = self.driver.size(self.get_path())
        if 'create' in dir(self.driver) and not blocks:
            return 0
        if blocks is None:
            return None  # Don't know
        return blocks  # Existing size
    def get_virtual_size(self):
        blocks = self.driver.vsize(self.get_path())
        if 'create' in dir(self.driver) and not blocks:
            return self._blocks  # Will create this size
        if blocks is None:
            return None  # Don't know
        return blocks  # Existing size
    def set_virtual_size(self, blocks):
        self._blocks = None  # Can't set
        if blocks is None:
            return
        assert(type(blocks) in (int, long))
        if 'create' in dir(self.driver) and not self.driver.size(self.get_path()):
            self._blocks = blocks

    def _update_template_path(self):
        if self._template_lock:
            return
        num = -1
        while True:
            num += 1
            used = False
            path = os.path.join(vminstall.paths.imagedir, self.vmname, 'disk%d' % num)
            if os.path.exists(path):
                continue
            for disk in self.disk_group:
                if disk is self:
                    continue
                has_path = disk._has_path()
                if has_path and has_path == path:
                    used = True
                    break
            if not used:
                break
        self._template_path = path
    def _has_path(self):
        return (self._template and self._template_path) or self._path
    def get_proto(self):
        return self.driver.proto
    def get_path(self):
        if self._template:
            if not self._template_path:
                self._update_template_path()
            return self._template_path
        return self._path
    def get_pdev(self):
        return '%s:%s' % (self.driver.proto, self.get_path())
    def set_pdev(self, pdev):
        assert(not self._template_lock and not self._block_vdev and not self._created)
        self._template_path = None
        if pdev:
            self.driver, self._path = split_driver_path(pdev)
            self._template = False
        else:
            self.driver = DRIVERS[default_proto]
            self._template = True
            self._path = None
    pdev = property(get_pdev, set_pdev)

    def get_ro(self):
        return self._ro
    def set_ro(self, ro):
        if self._vdevType == self.DEVICE_CDROM:
            ro = True
        elif self.exists() and self.driver.is_local():
            try:
                if (os.stat(self.get_path()).st_mode & stat.S_IWRITE) == 0:
                    ro = True
            except:
                pass
        if type(ro) is not bool:
            ro = False
        self._ro = ro
    ro = property(get_ro, set_ro)

    def get_xml_config(self):
        opts = []
        vals = self.driver.proto.split(':')
        if len(vals) == 1:
            opts.append("<driver name='%s'/>" % vals[0])
        else:
            opts.append("<driver name='%s' type='%s'/>" % (vals[0], vals[1]))
        opts.append("<source file='%s'/>" % self.get_path())
        opts.append("<target dev='%s'/>" % self.vdev)
        if self._ro:
            opts.append('<readonly/>')
        return [ "<disk type='%s' device='%s'>" % (self.driver.proto, self._vdevType),
                 opts,
                 '</disk>' ]
        
    def get_xen_config(self):
        mode = 'w'
        if self._ro:
            mode = 'r'
        vdev = self.vdev
        if self._vdevType == self.DEVICE_CDROM:
            vdev += ':cdrom'
            mode = 'r'
        return "'%s,%s,%s'" % (self.get_pdev(), vdev, mode)

    def __str__(self):
        size = ''
        blocks = self.get_virtual_size()
        if blocks is not None:
            size = '%1.1f GB ' % (blocks / (2.0*1024*1024))
        return '%s%s (%s)' % (size, self.get_vdev_type_str(), self.get_pdev())

    def __repr__(self):
        # PDEV,VDEV,TYPE,MODE,MB,DESCR
        ro = 'w'
        if self.get_ro():
            ro = 'r'
        blocks = self.get_virtual_size()
        if blocks is None:
            mb = ''
        else:
            mb = blocks / (2.0*1024)
        s = '%s,%s,%s,%s,%s,%s' % (self.get_pdev(), self.get_vdev_index(),
                                   self.get_vdevtype(), ro, mb,
                                   self.description)
        return s

    def _get_available_space_on_device(self, device_name=None):
        if device_name is None:
            device_name = self.get_path()
            if device_name is None:
                return -1
            elif device_name == "":
                device_name = os.getcwd()

        # Make sure pathname really exists
        pathname = device_name
        while True:
            if os.path.exists(pathname) or pathname == "/":
                break
            pathname = os.path.dirname(pathname)
            if pathname == "":
                return -1
        if pathname:
            device_name = pathname

        try:
            # Execute a df command.  Capture and parse the results
            df_output = commands.getoutput("/bin/df -B M %s" % device_name)
            if df_output:
                space_on_device = 0
                index = df_output.rfind('\n')
                last_line = df_output[index:].split()
                for s in last_line:
                    if s.endswith('M'):
                       space_on_device = s.rstrip('M')
                if space_on_device:
                    return (int(space_on_device) * (1024 * 1024)) / 512
            return -1
        except:
            return -1
