# Paravirtualized guest support
#
# Initially derived from virt-inst, Copyright 2005-2006 Red Hat, Inc.
# Copyright 2006-2007 Novell, Inc.
# Author: Charles Coffing <ccoffing@novell.com>
#
# This software may be freely redistributed under the terms of the GNU
# general public license.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import os
import re

from VMDefaults import VMDefaults
from XenGuest import *
from exceptions import *
from log import log
import mount
import paths
import util
import vmdisks

def _get_partitions(dev):
    """Returns a list of dictionaries that describe the (relevant)
    partitions on the device.
    """
    partitions = []
    pre = re.compile('([^ ]*)[ \\t]*:[ \\t]*start=[ \\t]*([0-9]*)[ \\t]*, size=[ \\t]*([0-9]*)[ \\t]*, Id=[ \\t]*([0-9a-fA-F]*)')
    log.debug("Examining '%s' for partitions..." % dev)
    lines = util.run(['/sbin/sfdisk', '--force', '-d', '-uS', '--', dev])
    for line in lines.splitlines():
        m = pre.match(line)
        if m:
            partition = {}
            pname = m.group(1)
            pstart = m.group(2)
            pid = m.group(4)
            # For now, only investigate 83 (Linux) and 6 (FAT16)
            if pid not in ('6', '83'):
                continue
            log.debug("Found partition '%s'." % (pname,))
            partition['name'] = pname
            partition['id'] = int(pid)
            partition['offset'] = int(pstart) * 512
            partitions.append(partition)
    return partitions

class ParaVirtGuest(XenGuest):
    def __init__(self, options):
        # Set these early, in case the XenGuest.__init__ raises.
        self.kernel = ''
        self.initrd = ''
        XenGuest.__init__(self, options)
        self.runtime = {}

    def _get_paravirt_images(self, dev, partitions, copy, progress=lambda f:None):
        def traildigits(strg):
            """Return the trailing digits, used to split the partition number off"""
            idx = len(strg)-1
            while strg[idx].isdigit():
                if len == 0:
                    return strg
                idx -= 1
            return strg[idx+1:]
        actions = VMDefaults[self.options.os_type].actions()
        for partition in partitions:
            # For now, only investigate 6 (FAT16) and 83 (Linux)
            if partition['id'] not in (6, 83):
                continue
            log.debug("Loop-mounting '%s', offset %d..." % (dev, partition['offset']))
            vdisk = mount.TemporaryLoopMount(dev)
            try:
                mountpoint = vdisk.mount(options='loop,ro,offset=%d' % partition['offset'])
            except Exception, e:
                log.debug("Mount failed (%s); skipping." % str(e))
                continue
            try:
                (_kernel, _initrd) = actions.getRuntimeImagesFromPath(mountpoint)
                if _kernel:
                    vpart = traildigits(partition['name'])
                    log.info("Found kernel on partition %s." % vpart)
                    if copy:
                        # Be careful to explicitly close on failure, and close
                        # immediately after use, otherwise no guarantee these
                        # will be closed by GC before attempting unmount.
                        kfd = open(mountpoint + _kernel, 'r')
                        try:
                            ifd = None
                            if _initrd:
                                ifd = open(mountpoint + _initrd, 'r')
                        except:
                            kfd.close()
                            raise
                        try:
                            kernel = util.copy_to_temp(kfd, prefix='kernel.')
                            if not ifd:
                                initrd = None
                            else:
                                try:
                                    initrd = util.copy_to_temp(ifd, prefix='initrd.')
                                except:
                                    try: os.unlink(kernel)
                                    except: pass
                                    raise
                        finally:
                            kfd.close()
                            if ifd:
                                ifd.close()
                    else:
                        log.info("Not copying; will use kernel/initrd in-place")
                        kernel = _kernel
                        initrd = _initrd
                    log.info("kernel %s; initrd %s" % (kernel, initrd))
                    return (vpart, kernel, initrd)
            finally:
                vdisk.unmount()
                progress(1.0)
        progress(1.0)
        return (None, None, None)

    def _get_paravirt_install_images(self, progress=lambda f:None):
        """Retrieves the kernel and initrd from the installation source.
        Potentially a heavy-weight operation."""
        actions = VMDefaults[self.options.os_type].actions()
        source = self.options.source
        nonpae = VMDefaults[self.options.os_type].supports_32nonpae()
        if type(source) is vmdisks.Disk:
            log.debug("Installation source: %s" % source)
            if not source.exists():
                raise InstSrcError(InstSrcError.E_NOT_FOUND, details=str(source))
            vdev = source.block_attach()
            try:
                partitions = _get_partitions(vdev)
                if len(partitions) == 0:
                    mounted_disk = mount.TemporaryLoopMount(vdev)
                    try:
                        mountpoint = mounted_disk.mount()
                    except Exception, e:
                        raise InstSrcError(InstSrcError.E_BAD, details=e.args[0])
                    try:
                        (kfn, ifn) = actions.getInstallImagesFromPath(mountpoint, progress, nonpae)
                    finally:
                        mounted_disk.unmount()
                else:
                   (_, kfn, ifn) = self._get_paravirt_images(vdev, partitions, copy=True, progress=progress)
            finally:
                source.block_detach()
        else:
            (kfn, ifn) = actions.getInstallImagesFromUrl(source, progress, nonpae)
        if kfn is None:
            raise InstSrcError(InstSrcError.E_BAD, details=str(source))
        return (kfn, ifn)

    def _get_install_xml(self):
        os = "<os type='%s'" % self.options.os_type
        if self.options.install:
            if type(self.options.source) is vmdisks.Disk:
                src = 'dev:/' + self.options.source.get_vdev_str()
            else:
                src = self.options.source
            os += " source='%s'" % src
        os += '>'
        return [ os,
                 [ '<type>linux</type>',
                   '<kernel>%s</kernel>' % self.kernel,
                   '<initrd>%s</initrd>' % self.initrd,
                   '<cmdline>%s %s</cmdline>' % (self.options._install_args, self.options.extra_args) ],
                 '</os>' ]

    def _get_runtime_xml(self):
        ramdisk = ''
        if self.runtime['initrd']:
            ramdisk = ',%s' % self.runtime['initrd']
        return [ "<os type='%s'>" % self.options.os_type,
                 [ '<type>linux</type>',
                   '<bootloader>%s</bootloader>' % self.options.loader,
                   '<bootargs>--entry=%s:%s%s</bootargs>' % (self.runtime['vdev'], self.runtime['kernel'], ramdisk),
                   '<cmdline>%s %s</cmdline>' % (self.options._runtime_args, self.options.extra_args) ],
                 '</os>' ]

    def GetConfigXml(self):
        if self.isInstall:
            os_xml = self._get_install_xml()
        else:
            os_xml = self._get_runtime_xml()
            
        return [ "<domain type='xen'>",
                 self.getCommonXml(),
                 os_xml,
                 [ '<devices>',
                   [ self._get_disk_xml(),
                     self._get_network_xml(),
                     self._get_graphics_xml() ],
                   '</devices>' ],
                 '</domain>' ]
                 
    def _get_install_xen(self):
        config = 'builder="linux"\nkernel="%s"\n' % self.kernel
        if self.initrd:
            config += 'ramdisk="%s"\n' % self.initrd
        config += 'extra="%s %s"\n' % (self.options._install_args, self.options.extra_args)
        return config

    def _get_runtime_xen(self):
        ramdisk = ''
        if self.runtime['initrd']:
            ramdisk = ',%s' % self.runtime['initrd']
        return """builder="linux"
bootloader="%(loader)s"
bootargs="--entry=%(vdev)s:%(kernel)s%(ramdisk)s"
extra="%(runtime_args)s %(extra_args)s"
""" % { 'loader': self.options.loader, 'vdev': self.runtime['vdev'], 'kernel': self.runtime['kernel'],
        'ramdisk': ramdisk, 'runtime_args': self.options._runtime_args, 'extra_args': self.options.extra_args }

    def GetConfigXen(self):
        if self.isInstall:
            osblob = self._get_install_xen()
        else:
            osblob = self._get_runtime_xen()

        return """%(common)s%(osblob)s%(disks)s
%(networks)s
%(graphics)s
""" % { 'common': self.getCommonXen(), 'osblob': osblob, 'disks': self._get_disk_xen(), 'networks': self._get_network_xen(), 'graphics': self._get_graphics_xen() }

    def PreInstall(self, progress=lambda f:None):
        XenGuest.PreInstall(self, lambda f:progress(f*0.5))
        self._cleanup_kernel()  # In case the install was retried
        (self.kernel, self.initrd) = self._get_paravirt_install_images(lambda f:progress(0.5+f*0.5))

    def Start(self):
        try:
            XenGuest.Start(self)
        finally:
            self._cleanup_kernel()
        
    def PostInstallCheck(self):
        """Probes the installed disk(s) for kernel/initrd to boot from."""
        def _get_paravirt_runtime_images():
            r = (None, None, None)
            for disk in self.options.disks:
                # FIXME: skip installation source?
                try:
                    vdev = disk.block_attach()
                except:
                    continue
                try:
                    try:
                        partitions = _get_partitions(vdev)
                        r = self._get_paravirt_images(vdev, partitions, copy=False)
                    except:
                        continue
                finally:
                    disk.block_detach()
                    if r[0] or r[1]:
                        break
            if r[0] is None or r[1] is None:
                raise VmError(VmError.E_NO_KERNEL)
            return (disk.get_vdev_str() + r[0], r[1], r[2])
        (vdev, kernel, initrd) = _get_paravirt_runtime_images()
        # Success:  No turning back to retry installation now!
        self.runtime['vdev'] = vdev
        self.runtime['kernel'] = kernel
        self.runtime['initrd'] = initrd

    def _cleanup_kernel(self):
        if self.kernel:
            try:
                os.unlink(self.kernel)
            finally:
                self.kernel = ''
        if self.initrd:
            try:
                os.unlink(self.initrd)
            finally:
                self.initrd = ''

    def __del__(self):
        self._cleanup_kernel()
