# Default VM settings for various OSs
#
# Copyright 2006-2007 Novell, Inc.
# Author: Charles Coffing <ccoffing@novell.com>
#
# This software may be freely redistributed under the terms of the GNU
# general public license.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

from nics import XenNic
from VMActions import VMActions
from VMNetWare import VMNetWare
from VMRedHat import VMRedHat
from VMSUSE import VMSUSE
from VMWindows import VMWindows
from vmdisks import Disk
import caps
import msg
import paths


class GenericDefaults:
    def group(self):
        return 'other'
    def actions(self):
        return VMActions()
    def install_sources(self):
        return ('generic', ['dev:/'])
    def name(self):
        return msg.other_os
    def virt_support(self):
        return ('xen', 'hvm')
    def apic(self):
        return False
    def max_cpus(self):
        return 32
    def default_memory(self):
        return 256
    def min_memory(self):
        return 4
    def max_memory(self):
        if caps.is_x86_64_capable():
            mb = 128*1024  # Arbitrarily large...
        else:
            if caps.is_pae_capable():
                mb = 16*1024  # PAE limits at 64GB; Xen has lower per-vm limits
            else:
                mb = 2*1024  # hardware limit is 4GB; Xen has lower per-vm limits
        return mb
    def _disk_template(self, num, sizeGB):
        return Disk(pdev=None, vdev=num, blocks=sizeGB*1024*1024*2, vdevType=Disk.DEVICE_DISK)
    def _cdrom_template(self, num):
        return Disk(pdev=None, vdev=num, vdevType=Disk.DEVICE_CDROM)
    def _disk_size(self):
        return 4
    def disks(self):
        return [ self._disk_template(0, self._disk_size()) ]
    def nics(self):
        return [ XenNic() ]
    def localtime(self):
        return False
    def xterm_args(self):
        return ['+u8', '-fg', 'white', '-bg', 'black']
    def auto_install(self):
        return None
    def num_vcpus(self):
        return 1
    def boot_pxe(self):
        return False
    def supports_32nonpae(self):
        return True

class LinuxDefaults(GenericDefaults):
    def xterm_args(self):
        # Linux behaves much better with an 80x24 xterm.
        args = GenericDefaults.xterm_args(self)
        args.extend(['-geometry', '80x24'])
        return args

class RedHatDefaults(LinuxDefaults):
    def group(self):
        return 'redhat'
    def name(self):
        return msg.redhat_other
    def install_sources(self):
        return ['redhat', ('ftp://', 'http://', 'nfs://', 'dev:/')]
    def actions(self):
        return VMRedHat()
    def virt_support(self):
        return ('xen', 'hvm')
    def min_memory(self):
        # Minimum for text-based install
        # As per http://fedora.redhat.com/docs/release-notes/
        return 128
    def default_memory(self):
        # FIXME:  how do I know if install source is 32 or 64 bit?
        #if caps.is_x86_64_capable():
        #    return 512
        #else:
        return 256
    def auto_install(self):
        return msg.kickstart_desc
    def apic(self):
        # RHEL3 64bit needs this.  Slows down 32bit:(
        return True

# https://www.redhat.com/rhel/details/limits/
class RHEL3Defaults(RedHatDefaults):
    def name(self):
        return msg.rhel3
    def virt_support(self):
        return ('hvm')
    def max_cpus(self):
        return 16
    def min_memory(self):
        return 256
    def max_memory(self):
        return 8*1024

class RHEL4Defaults(RedHatDefaults):
    def name(self):
        return msg.rhel4
    def virt_support(self):
        # RHEL4 doesn't support para, but RHEL4 update 8 does, so prefer hvm.
        return ('hvm', 'xen')
    def min_memory(self):
        return 256
    def max_memory(self):
        return 16*1024
    def _disk_size(self):
        return 8
    def disks(self):
        return [ self._disk_template(0, self._disk_size()) ]
    def num_vcpus(self):
        return caps.vcpu_processors(4);

class RHEL5Defaults(RedHatDefaults):
    def name(self):
        return msg.rhel5
    def min_memory(self):
        return 256
    def max_memory(self):
        return 16*1024
    def _disk_size(self):
        return 8
    def disks(self):
        return [ self._disk_template(0, self._disk_size()) ]
    def num_vcpus(self):
        return caps.vcpu_processors(4);

class SUSEDefaults(LinuxDefaults):
    def group(self):
        return 'suse'
    def name(self):
        return msg.suse_other
    def virt_support(self):
        return ('xen', 'hvm')
    def install_sources(self):
        return ['suse', ('ftp://', 'http://', 'nfs://', 'smb://', 'dev:/')]
    def actions(self):
        return VMSUSE()
    def min_memory(self):
        # Minimum for text-based install
        return 192
    def default_memory(self):
        return 256
    def auto_install(self):
        return msg.autoyast_desc

class openSUSEDefaults(SUSEDefaults):
    def name(self):
        return msg.opensuse
    def default_memory(self):
        return 384
    def _disk_size(self):
        return 8
    def disks(self):
        return [ self._disk_template(0, self._disk_size()) ]

class openSUSE11Defaults(SUSEDefaults):
    def name(self):
        return msg.opensuse11
    def default_memory(self):
        return 384
    def _disk_size(self):
        return 8
    def disks(self):
        return [ self._disk_template(0, self._disk_size()) ]
    def supports_32nonpae(self):
        return False

class SLED10Defaults(SUSEDefaults):
    def name(self):
        return msg.sled10
    def min_memory(self):
        return 256
    def default_memory(self):
        return 512

class SLED11Defaults(SUSEDefaults):
    def name(self):
        return msg.sled11
    def min_memory(self):
        return 256
    def default_memory(self):
        return 512
    def supports_32nonpae(self):
        return False

class SLES8Defaults(SUSEDefaults):
    def name(self):
        return msg.sles8
    def virt_support(self):
        return ('hvm')

class SLES9Defaults(SUSEDefaults):
    def name(self):
        return msg.sles9
    def default_memory(self):
        return 512
    def virt_support(self):
        # SLES9 supports para starting with SP4
        return ('hvm', 'xen')
    def apic(self):
        # Turned on even for uni-processor
        return True
    def num_vcpus(self):
        return caps.vcpu_processors(2);
    def apic(self):
        return True

class SLES10Defaults(SUSEDefaults):
    def name(self):
        return msg.sles10
    def default_memory(self):
        return 512
    def num_vcpus(self):
        return caps.vcpu_processors(4);
    def apic(self):
        return True

class SLES11Defaults(SUSEDefaults):
    def name(self):
        return msg.sles11
    def default_memory(self):
        return 512
    def num_vcpus(self):
        return caps.vcpu_processors(4);
    def apic(self):
        return True
    def supports_32nonpae(self):
        return False

class NetWareDefaults(GenericDefaults):
    def group(self):
        return 'netware'
    def name(self):
        return msg.nw65_older
    def virt_support(self):
        return ('hvm')
    def actions(self):
        return VMNetWare()
    def min_memory(self):
        return 32
    def default_memory(self):
        return 512
    def localtime(self):
        return True
    def xterm_args(self):
        # NetWare behaves much better with an 80x25 xterm.
        args = GenericDefaults.xterm_args(self)
        args.extend(['-geometry', '80x25'])
        return args
    def auto_install(self):
        return msg.nwresponse_desc
    def num_vcpus(self):
        return caps.vcpu_processors(2);

class NOES2NWDefaults(NetWareDefaults):
    def name(self):
        return msg.oes2nw
    def virt_support(self):
         return ('xen')
    def min_memory(self):
        return 512
    def default_memory(self):
        return 512
    def _disk_size(self):
        # Docs say minimum 6, recommended 10
        return 10
    def disks(self):
        return [ self._disk_template(0, self._disk_size()) ]

class NOES2LDefaults(SUSEDefaults):
    def name(self):
        return msg.oes2l
    def min_memory(self):
        return 512
    def default_memory(self):
        return 512
    def _disk_size(self):
        # Docs say minimum 6, recommended 10
        return 10
    def disks(self):
        return [ self._disk_template(0, self._disk_size()) ]
    def num_vcpus(self):
        return caps.vcpu_processors(4);

class WindowsDefaults(GenericDefaults):
    def group(self):
        return 'windows'
    def virt_support(self):
        return ('hvm')
    def actions(self):
        return VMWindows()
    def localtime(self):
        return True
    def _disk_size(self):
        return 8
    def disks(self):
        return [ self._disk_template(0, self._disk_size()) ]
    def min_memory(self):
        return 64
    def default_memory(self):
        return 256
class WindowsNTDefaults(WindowsDefaults):
    def name(self):
        return msg.windows_nt
    def disks(self):
        return [ self._disk_template(0, self._disk_size()) ]
    def min_memory(self):
        return 16
    def default_memory(self):
        return 128
    def max_memory(self):
        # WARNING:  NT has installer bug; crashes if > 256 MB RAM; can
        #           manually increase post-install.
        return 256
class WindowsOtherDefaults(WindowsDefaults):
    def name(self):
        return msg.windows_other
class WindowsOtherx64Defaults(WindowsDefaults):
    def name(self):
        return msg.windows_other_x64
    def apic(self):
        return True
    def num_vcpus(self):
        return caps.vcpu_processors(2);
class WindowsXPDefaults(WindowsDefaults):
    def name(self):
        return msg.windows_xp
    def default_memory(self):
        return 512
    def num_vcpus(self):
        return caps.vcpu_processors(2);
    def apic(self):
        return True
class WindowsXPx64Defaults(WindowsDefaults):
    def name(self):
        return msg.windows_xp_x64
    def default_memory(self):
        return 512
    def num_vcpus(self):
        return caps.vcpu_processors(2);
    def apic(self):
        return True
class WindowsVistaDefaults(WindowsDefaults):
    def name(self):
        return msg.windows_vista
    def min_memory(self):
        return 512
    def default_memory(self):
        return 1024
    def _disk_size(self):
        return 20
    def disks(self):
        return [ self._disk_template(0, self._disk_size()) ]
    def num_vcpus(self):
        return caps.vcpu_processors(2);
    def apic(self):
        return True
class WindowsVistax64Defaults(WindowsVistaDefaults):
    def name(self):
        return msg.windows_vista_x64
class WindowsServer2008Defaults(WindowsDefaults):
    def name(self):
        return msg.windows_server_2008
    def virt_support(self):
        # 'hvm-ext' here means enlightened windows - AKA, Viridian
        return ('hvm-ext', 'hvm')
    def min_memory(self):
        return 512
    def default_memory(self):
        return 1024
    def _disk_size(self):
        return 40
    def disks(self):
        return [ self._disk_template(0, self._disk_size()) ]
    def enlightened():
        return True
    def num_vcpus(self):
        return caps.vcpu_processors(4);
    def apic(self):
        return True
class WindowsServer2008x64Defaults(WindowsServer2008Defaults):
    def name(self):
        return msg.windows_server_2008_x64

class Solaris9Defaults(GenericDefaults):
    def group(self):
        return 'solaris'
    def name(self):
        return msg.solaris9
    def virt_support(self):
        return ('hvm')
    # http://www.sun.com/software/solaris/9/specs.xml
    def min_memory(self):
        return 64
    def default_memory(self):
        return 128
    def num_vcpus(self):
        return caps.vcpu_processors(4);
    def apic(self):
        return True
# FIXME:  learn about this...
#    def auto_install(self):
#        return "JumpStart"
class Solaris10Defaults(Solaris9Defaults):
    def name(self):
        return msg.solaris10
    # http://www.sun.com/software/solaris/specs.jsp
    def min_memory(self):
        return 256
    def default_memory(self):
        return 512
    def virt_support(self):
        # FIXME:  need to code support for bootstrapping PV Solaris!
        # http://www.opensolaris.org/os/community/xen/
        return ('hvm', 'xen')


VMDefaults = {
    'netware'  : NetWareDefaults(),
    'oes2l'    : NOES2LDefaults(),
    'oes2nw'   : NOES2NWDefaults(),
    'opensuse11' : openSUSE11Defaults(),
    'opensuse' : openSUSEDefaults(),
    'other'    : GenericDefaults(),
    'redhat'   : RedHatDefaults(),
    'rhel3'    : RHEL3Defaults(),
    'rhel4'    : RHEL4Defaults(),
    'rhel5'    : RHEL5Defaults(),
    'sled10'   : SLED10Defaults(),
    'sled11'   : SLED11Defaults(),
    'sles10'   : SLES10Defaults(),
    'sles11'   : SLES11Defaults(),
    'sles8'    : SLES8Defaults(),
    'sles9'    : SLES9Defaults(),
    'suse'     : SUSEDefaults(),
    'solaris10': Solaris10Defaults(),
    'solaris9' : Solaris9Defaults(),
    'winserver2008'   : WindowsServer2008Defaults(),
    'winserver2008x64': WindowsServer2008x64Defaults(),
    'windowsvista'   : WindowsVistaDefaults(),
    'windowsvistax64': WindowsVistax64Defaults(),
    'windows'        : WindowsOtherDefaults(),
    'windowsx64'     : WindowsOtherx64Defaults(),
    'windowsxp'      : WindowsXPDefaults(),
    'windowsxpx64'   : WindowsXPx64Defaults(),
    'windowsnt'      : WindowsNTDefaults(),
}

default_os_type = 'opensuse11'
