# Common code for all Xen guests
#
# Initially derived from virt-inst, Copyright 2005-2006 Red Hat, Inc.
# Copyright 2006-2007 Novell, Inc.
# Author: Charles Coffing <ccoffing@novell.com>
#
# This software may be freely redistributed under the terms of the GNU
# general public license.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import os

from xen.xend import XendOptions
from VMDefaults import VMDefaults
from exceptions import *
from log import log
import paths
import util
import vmdisks
import xend
import keytable
import re

KEYBOARD_DIR = "/etc/sysconfig/keyboard"

class XenGuest(object):
    def __init__(self, options):
        self.options = options
        self.uuid = None          # None if not running else == self.options.uuid
        self.isInstall = True
        self.tmpdisks = []

    def os(self):
        return VMDefaults[self.options.os_type]
        
    def _get_name_xml(self):
        return ['<name>%s</name>' % self.options.name]
    def _get_name_xen(self):
        return 'name="%s"\n' % self.options.name

    def _get_memory_xml(self):
        # XML wants KB
        return ['<memory>%d</memory>' % (self.options.memoryMB*1024,),
                '<maxmemory>%d</maxmemory>' % (self.options.max_memoryMB*1024,)]
    def _get_memory_xen(self):
        # config file wants MB
        # FIXME:  xen's "maxmem" setting is broken -- debug xen before setting
        return 'memory=%d\n' % self.options.memoryMB

    def _get_uuid_xml(self):
        return ['<uuid>%s</uuid>' % self.options.uuid]
    def _get_uuid_xen(self):
        return 'uuid="%s"\n' % self.options.uuid

    def _get_action(self, action):
        if self.isInstall:
            return 'destroy'
        else:
            if action == 'crash': return self.options.on_crash
            if action == 'poweroff': return self.options.on_poweroff
            if action == 'reboot': return self.options.on_reboot
    def _get_action_xml(self):
        return ['<on_poweroff>%s</on_poweroff>' % self._get_action('poweroff'),
                '<on_reboot>%s</on_reboot>' % self._get_action('reboot'),
                '<on_crash>%s</on_crash>' % self._get_action('crash')]
    def _get_action_xen(self):
        return 'on_poweroff="%s"\non_reboot="%s"\non_crash="%s"\n' % \
               (self._get_action('poweroff'), self._get_action('reboot'), self._get_action('crash'))

    def _get_vcpus_xml(self):
        return ['<vcpu>%d</vcpu>' % self.options.vcpus]
    def _get_vcpus_xen(self):
        return 'vcpus=%d\n' % self.options.vcpus

    def _get_localtime_xml(self):
        if self.os().localtime():
            return ["<clock sync='localtime'/>"]
        return ["<clock sync='utc'/>"]
    def _get_localtime_xen(self):
        if self.os().localtime():
            return 'localtime=1\n'
        return 'localtime=0\n'

    def getCommonXen(self):
        s = ''
        s += self._get_name_xen()
        s += self._get_uuid_xen()
        s += self._get_memory_xen()
        s += self._get_vcpus_xen()
        s += self._get_action_xen()
        s += self._get_localtime_xen()
        s += self._get_keymap_xen()
        return s

    def getCommonXml(self):
        s = []
        s.extend(self._get_name_xml())
        s.extend(self._get_uuid_xml())
        s.extend(self._get_memory_xml())
        s.extend(self._get_vcpus_xml())
        s.extend(self._get_action_xml())
        s.extend(self._get_localtime_xml())
        s.extend(self._get_keymap_xml())
        return s

    def _get_disk_xml(self):
        """Get the disk config in the libvirt XML format"""
        ret = []
        for disk in self.options.disks + self.tmpdisks:
            ret.extend(disk.get_xml_config())
        return ret

    def _get_disk_xen(self):
        """Get the disk config in the xend python format"""
        ret = 'disk=[ '
        for disk in self.options.disks + self.tmpdisks:
            ret += disk.get_xen_config()
            ret += ', '
        ret += ']'
        return ret

    def _get_network_xml(self):
        """Get the network config in the libvirt XML format"""
        nics = []
        for n in self.options.nics:
            model = ""
            if n.model:
                model = " model='%s'" % n.model
            nic = ["<interface type='bridge'%s>" % model]
            opts = []
            if n.bridge:
                opts.append("<source bridge='%s'/>" % n.bridge)
            if n.mac:
                opts.append("<mac address='%s'/>" % n.mac)
            opts.append("<script path='/etc/xen/scripts/vif-bridge'/>")
            nic.append(opts)
            nic.append('</interface>')
            nics.append(nic)
        return nics

    def _get_network_xen(self):
        """Get the network config in the xend python format"""
        ret = 'vif=[ '
        for nic in self.options.nics:
            ret += "'"
            needComma = 0
            commas = [ '', ',' ]
            if nic.mac:
                ret += commas[needComma]
                ret += 'mac=%s' % (nic.mac,)
                needComma = 1
            if nic.bridge:
                ret += commas[needComma]
                ret += 'bridge=%s' % ( nic.bridge, )
                needComma = 1
            if self.options.full_virt and nic.model != 'para':
                ret += commas[needComma]
                if nic.model:
                    ret += 'model=%s,' % ( nic.model, )
                ret += 'type=ioemu'
                needComma = 1
            ret += "', "
        ret += ']'
        return ret

    def _get_keymap(self):
        keymap = None
        if self.options.keymap != None:
            return self.options.keymap
        if XendOptions.instance().get_keymap():
            return None
        try:
            f = open(KEYBOARD_DIR, "r")
        except IOError, e:
            logging.debug('Could not open "/etc/sysconfig/keyboard" ' + str(e))
        else:
            while 1:
                s = f.readline()
                if s == "":
                    break
                s = re.sub("\s+", "", s)
                if re.search('^KEYTABLE=\".*\"$', s) != None:
                        kt = s.split('"')[1]
                        if ".map.gz" in kt:
                            kt = kt.rstrip(".map.gz") 
                        if keytable.keytable.has_key(kt):
                            keymap = keytable.keytable[kt]
            f.close
        return keymap;

    def _get_keymap_xml(self):
        ret = []
        keymap = self._get_keymap()
        if keymap != None:
           ret = ['<keymap>%s</keymap>' % keymap]
        return ret

    def _get_keymap_xen(self):
        ret = ''
        keymap = self._get_keymap()
        if keymap != None:
           ret = "keymap=\"%s\"\n" %(keymap)
        return ret

    def _get_graphics_xml(self):
        """Get the graphics config in the libvirt XML format."""
        if self.options.graphics == 'none':
            return []
        ret = ''
        if self.options.graphics_viewer == 'vnc':
            ret = "<graphics type='vnc'"
            if self.options.vncport:
                ret += " port='%d'" %(self.options.vncport,)
            if self.options.vnc_password:
                ret += " password='%s'" %(self.options.vnc_password,)
            ret += '/>'
        elif self.options.graphics_viewer == 'sdl':
            ret = "<graphics type='sdl'/>"
        return [ret]

    def _get_graphics_xen(self):
        """Get the graphics config in the xend python format"""
        if self.options.graphics == 'none':
            return 'nographic=1'

        lines = []
        options = []
        viewer = self.options.graphics_viewer
        if viewer == 'vnc':
            if self.options.vnc_password:
                options.append("vncpasswd='%s'" % self.options.vnc_password)
            port = self.options.vncport
            if port is None:
                options.append('vncunused=1')
            elif port >= 5900:
                options.append('vncdisplay=%d' % (port - 5900,))
                options.append('vncunused=0')
        else:
            assert(viewer == 'sdl')
            options=[]

        if self.options.graphics == 'para':
            options = map(lambda s: "".join(s.split("'")), options)
            lines.append('vfb=[\'type=%s,%s\']' % (viewer, ','.join(options)))
        else:
            stdvga = 0
            if self.options.graphics == 'vesa':
                stdvga=1
            lines.append('stdvga=%d' % stdvga)
            lines.append('%s=1' % viewer)
            lines += options

        return '\n'.join(lines)

    def GetConfigXml(self):
        raise RuntimeError('pure virtual')
    def GetConfigXen(self):
        raise RuntimeError('pure virtual')

    def PreInstall(self, progress=lambda f:None):
        if xend.connection.domain_running(self.options.name):
            raise VmError(VmError.E_VM_EXISTS)
        if self.options.os_settings:
            self.os().actions().setupOsSettings(self, lambda f:progress(f*0.05))
        self.create_devices(lambda f:progress(0.05 + f*0.95))

    def Start(self):
        if (xend.connection.domain_running(self.options.name) or
            xend.connection.domain_running(self.options.uuid)):
            raise VmError(VmError.E_VM_EXISTS)
        # If there is an existing configuration by this name it is not running,
        # but it may have a different (conflicting) UUID, so delete it.
        xend.connection.delete(self.options.name)
        try:
            util.run(['/usr/sbin/xm', 'new', os.path.join(paths.configdir, self.options.name)])
            util.run(['/usr/sbin/xm', 'start', self.options.name])
        except RuntimeError, e:
            log.info("'xm new' or 'xm start' failed")
            self.Stop()
            xend.connection.delete(self.options.name)
            raise VmError(VmError.E_NO_START, details=e.args[0])
        if not xend.connection.wait_domain(self.options.name):
            log.info("Domain failed to start")
            self.Stop()
            xend.connection.delete(self.options.name)
            raise VmError(VmError.E_NO_START)
        self.uuid = self.options.uuid

    def Stop(self):
        # Slightly safer to stop by running-uuid, if it exists, rather than name.
        name = self.uuid
        if not name:
            name = self.options.name
        try:
            xend.connection.destroy(name)
        except:
            pass
        self.uuid = None

    def PostInstallConf(self):
        """Perform any post-install cleanup and configuration."""
        self.isInstall = False
        self._cleanup_tmpdisks()
        self.os().actions().postInstall(self)

    def manage_disks(self):
        dms = vmdisks.diskmap.DiskMapStatic()
        for disk in self.options.disks:
            dms.manage_disk(disk)
        dms.flush()

    def unmanage_disks(self):
        dms = vmdisks.diskmap.DiskMapStatic()
        for disk in self.options.disks:
            try:
                dms.unmanage_disk(disk)
            except:
                # Possible that someone rm'd the disk list file...
                pass
        dms.flush()

    def start_from_disk(self):
        """Restart the guest from its disks."""
        self.create_devices()
        # Ensure that the install settings are removed; overlaying settings
        # might break (specifically in the case of kernel vs. bootloader).
        xend.connection.delete(self.options.name)
        try:
            util.run(['/usr/sbin/xm', 'new', os.path.join(paths.configdir, self.options.name)])
            util.run(['/usr/sbin/xm', 'start', self.options.name])
        except RuntimeError, e:
            log.info("'xm new' or 'xm start' failed")
            self.Stop()
            xend.connection.delete(self.options.name)
            raise VmError(VmError.E_NO_START, details=e.args[0])
        if not xend.connection.wait_domain(self.options.name):
            log.info("Domain failed to start")
            self.Stop()
            xend.connection.delete(self.options.name)
            raise VmError(VmError.E_NO_START)
        self.uuid = self.options.uuid

    def create_devices(self, progress=lambda f:None):
        disks = self.options.disks + self.tmpdisks
        times = map(lambda d: d.setup_time(), disks)
        cum_times = [0.0]
        for i in range(len(times)):
            cum_times.append(cum_times[-1]+times[i])
        for i in range(len(disks)):
            disks[i].setup(lambda f: progress(cum_times[i]/cum_times[-1] + f*cum_times[i+1]/cum_times[-1]))

    def FailedInstall(self):
        log.info("Installation failed, so destroying VM")
        self.Stop()
        xend.connection.delete(self.options.name)
        log.info("Installation failed, so cleaning up disks")
        self._cleanup_tmpdisks()
        self.unmanage_disks()
        for disk in self.options.disks:
            disk.reset()

    def _cleanup_tmpdisks(self):
        for disk in self.tmpdisks:
            try:
                os.unlink(disk.get_path())
            except:
                pass
        self.tmpdisks = []

    def __del__(self):
        self._cleanup_tmpdisks()
