# Utility functions used for guest installation
#
# Initially derived from virt-inst, Copyright 2005-2006 Red Hat, Inc.
# Copyright 2006-2007 Novell, Inc.
# Author: Charles Coffing <ccoffing@novell.com>
#
# This software may be freely redistributed under the terms of the GNU
# general public license.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import os
import random
import re
from subprocess import *
import tempfile

from log import log

def run(args):
    try:
        log.debug("Running: %s" % str(args))
        p = Popen(args=args, stdin=None, stdout=PIPE, stderr=STDOUT, close_fds=True,
                  env={'PATH': '/sbin:/usr/sbin:/bin:/usr/bin', 'TERM':'vt100'})
        (pid, status) = os.waitpid(p.pid, 0)
        output = p.stdout.read()
        if status:
            raise RuntimeError(output)
    except OSError, e:
        raise RuntimeError(str(e))
    return output

def close_fds(fds=xrange(0,256)):
    for fd in fds:
        try:
            os.close(fd)
        except OSError:
            pass
    if 0 in fds:
        os.open('/dev/null', os.O_RDWR)
    if 1 in fds:
        os.open('/dev/null', os.O_RDWR)
    if 2 in fds:
        os.open('/dev/null', os.O_RDWR)

def daemonize(fds=xrange(0,256), cb=None):
    pid = os.fork()
    if pid == 0:
        os.setsid()
        pid = os.fork()
        if pid == 0:
            os.chdir('/')
            os.umask(022)
            if cb:
                cb()
            close_fds(fds)
            return True
        else:
            os._exit(0)
    else:
        return False

# Word-wraps the text, and handles unicode.
# See http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/148061
rx=re.compile(u"([\u2e80-\uffff])", re.UNICODE)
def wrap(text, width=79, encoding="utf8"):
     return reduce(lambda line, word, width=width: '%s%s%s' %
                (line,
                 [' ','\n', ''][(len(line)-line.rfind('\n')-1
                       + len(word.split('\n',1)[0] ) >= width) or
                      line[-1:] == '\0' and 2],
                 word),
                rx.sub(r'\1\0 ', unicode(text,encoding)).split(' ')
            ).replace('\0', '').encode(encoding)

# Numeric-aware sorting of strings
# Inspired by http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/135435
def _generate_index(str):
    """Splits a string into alpha and numeric elements, which
    is used as an index for sorting"""
    index = []
    def _append(fragment, alist=index):
        if fragment.isdigit():
            fragment = int(fragment)
        alist.append(fragment)
    prev_isdigit = str[0].isdigit()
    current_fragment = ''
    for char in str:
        curr_isdigit = char.isdigit()
        if curr_isdigit == prev_isdigit:
            current_fragment += char
        else:
            _append(current_fragment)
            current_fragment = char
            prev_isdigit = curr_isdigit
    _append(current_fragment)    
    return tuple(index)

def sort_tuples_by_nth(alist, nth=0):
    indices = map(_generate_index, [t[nth] for t in alist])
    decorated = zip(indices, alist)
    decorated.sort()
    return [item[1] for item in decorated]

def sort_numeric_aware(alist):
    indices = map(_generate_index, alist)
    decorated = zip(indices, alist)
    decorated.sort()
    return [item for index, item in decorated]

def sort_dict_keys_by_value(d):
    sorted = sort_tuples_by_nth(d.items(), 1)
    return [ item[0] for item in sorted ]

def filtered_progress(fn, step=0.01):
    """Returns a function which takes a fraction (0.0 .. 1.0) describing
    progress.  When the progress advances beyond a specified step size,
    the function calls 'fn' with the current progress."""
    def p(f):
        if f < p.f + step:
            return
        p.f = f
        fn(p.f)
    p.f = 0.0
    fn(p.f)
    return p

def copy_to_temp(fileobj, prefix, progress=lambda f:None):
    (fd, fn) = tempfile.mkstemp(prefix=prefix)
    log.debug("Copying to temporary file '%s'" % fn)
    read = size = 0
    try:
        size = os.fstat(fileobj.fileno()).st_size
    except:
        pass
    block_size = 65536
    while 1:
        try:
            buff = fileobj.read(block_size)
            if not buff:
                break
            os.write(fd, buff)
        except:
            os.close(fd)
            os.unlink(fn)
            raise
        read += len(buff)
        if read <= size:  # Some file objects can't accurately report size
            progress(float(read) / float(size))
    os.close(fd)
    return fn

def mke2fsImage(filename, progress=lambda f:None):
    args = ['/sbin/mkfs.ext2', '-F', '-m', '0', '-q', '--', filename]
    log.debug("Creating filesystem: %s" % str(args))
    run(args)
    progress(1.0)

def mkdosfsImage(filename, progress=lambda f:None):
    args = ['/sbin/mkdosfs', '--', filename]
    log.debug("Creating filesystem: %s" % str(args))
    run(args)
    progress(1.0)

# the following three functions are from xend/uuid.py and are thus
# available under the LGPL,
# Copyright 2005 Mike Wray <mike.wray@hp.com>
# Copyright 2005 XenSource Ltd
def randomUUID():
    """Generate a random UUID."""
    return [ random.randint(0, 255) for _ in xrange(0, 16) ]

def uuidToString(u):
    return '-'.join(['%02x' * 4, '%02x' * 2, '%02x' * 2, '%02x' * 2,
                     '%02x' * 6]) % tuple(u)

def uuidFromString(s):
    s = s.replace('-', '')
    return [ int(s[i : i + 2], 16) for i in xrange(0, 32, 2) ]
